@@ -1,5 +1,7 @@
 #!/usr/bin/env python
 
+from __future__ import annotations
+
 import gzip
 import hashlib
 import json
@@ -11,7 +13,7 @@ from functools import wraps, lru_cache
 from io import BytesIO
 from operator import itemgetter
 from pathlib import Path
-from typing import Dict, Any, List, Optional, Tuple, Set, Union, Callable
+from typing import Any, Callable
 from urllib.parse import unquote_plus, urlparse
 
 from .helper import rebuild_url, Har2TreeError, Har2TreeLogAdapter
@@ -56,7 +58,7 @@ def trace_make_subtree_fallback(method: Callable[..., None]) -> Callable[..., No
 
 def trace_make_subtree(method: Callable[..., None]) -> Callable[..., None]:
     @wraps(method)
-    def _impl(self: Any, root: URLNode, nodes_to_attach: Optional[List[URLNode]]=None, dev_debug: bool=False) -> None:
+    def _impl(self: Any, root: URLNode, nodes_to_attach: list[URLNode] | None=None, dev_debug: bool=False) -> None:
         if dev_debug_mode:
             __load_debug_files()
             if dev_debug_url and root.name == dev_debug_url or nodes_to_attach is not None and any(True for u in nodes_to_attach if u.name == dev_debug_url):
@@ -92,18 +94,25 @@ class HarFile():
         self.path = harfile
 
         try:
-            self.har: Dict[str, Any]
+            self.har: dict[str, Any]
             if self.path.suffix == '.gz':
+                self.is_compressed = True
                 with gzip.open(self.path, 'rb') as f:
                     self.har = json.load(f)
             else:
+                self.is_compressed = False
                 with self.path.open() as f:
                     self.har = json.load(f)
         except json.decoder.JSONDecodeError as e:
             raise Har2TreeError(f'HAR file is not a valid JSON file: {e}')
 
+        root_name = self.path.stem
+        if self.is_compressed:
+            # we have .har.gz, and we want what comes *before* that
+            root_name = Path(root_name).stem
+
         # I mean it, that's the last URL the browser was on
-        last_redirect_file = self.path.parent / f'{self.path.stem}.last_redirect.txt'
+        last_redirect_file = self.path.parent / f'{root_name}.last_redirect.txt'
         if last_redirect_file.is_file():
             with last_redirect_file.open('r') as _lr:
                 self.final_redirect: str = unquote_plus(_lr.read())
@@ -112,17 +121,17 @@ class HarFile():
             self.logger.debug('No last_redirect file available.')
             self.final_redirect = ''
 
-        cookiefile = self.path.parent / f'{self.path.stem}.cookies.json'
+        cookiefile = self.path.parent / f'{root_name}.cookies.json'
         if cookiefile.is_file():
             with cookiefile.open() as c:
-                self.cookies: List[Dict[str, Any]] = json.load(c)
+                self.cookies: list[dict[str, Any]] = json.load(c)
         else:
             self.logger.debug('No cookies file available.')
             self.cookies = []
 
-        dlfile = self.path.parent / f'{self.path.stem}.data'
-        dlfilename = self.path.parent / f'{self.path.stem}.data.filename'
-        self.downloaded_file: Optional[BytesIO]
+        dlfile = self.path.parent / f'{root_name}.data'
+        dlfilename = self.path.parent / f'{root_name}.data.filename'
+        self.downloaded_file: BytesIO | None
         self.downloaded_filename: str
         if dlfile.is_file() and dlfilename.is_file():
             with dlfilename.open('r') as _fn:
@@ -134,8 +143,8 @@ class HarFile():
             self.downloaded_file = None
             self.downloaded_filename = ''
 
-        htmlfile = self.path.parent / f'{self.path.stem}.html'
-        self.html_content: Optional[BytesIO]
+        htmlfile = self.path.parent / f'{root_name}.html'
+        self.html_content: BytesIO | None
         if htmlfile.is_file():
             with htmlfile.open('rb') as _h:
                 self.html_content = BytesIO(_h.read())
@@ -149,12 +158,13 @@ class HarFile():
 
         # Used to find the root entry of a page in the capture
         # NOTE 2020-05-19: Turns out multiple pages can have the exact same timestamp...
-        self.pages_start_times: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
-        for page in self.har['log']['pages']:
-            self.pages_start_times[page['startedDateTime']].append(page)
-        # The first entry has a different start time as the one from the list, add that
-        if self.entries:
-            self.pages_start_times[self.initial_start_time].append(self.har['log']['pages'][0])
+        self.pages_start_times: dict[str, list[dict[str, Any]]] = defaultdict(list)
+        if 'pages' in self.har['log']:
+            for page in self.har['log']['pages']:
+                self.pages_start_times[page['startedDateTime']].append(page)
+            # The first entry has a different start time as the one from the list, add that
+            if self.entries:
+                self.pages_start_times[self.initial_start_time].append(self.har['log']['pages'][0])
 
         # Set to false if initial_redirects fails to find the chain.
         self.need_tree_redirects = False
@@ -181,7 +191,7 @@ class HarFile():
                 self.final_redirect = self.final_redirect.split('?', 1)[0]
                 self._search_final_redirect()
             else:
-                self.logger.warning(f'Unable to find the final redirect: {self.final_redirect}')
+                self.logger.info(f'Unable to find the final redirect: {self.final_redirect}')
 
     @property
     def number_entries(self) -> int:
@@ -191,7 +201,7 @@ class HarFile():
     @property
     def initial_title(self) -> str:
         """Title of the first page in the capture"""
-        if self.har['log']['pages'][0]['title']:
+        if 'pages' in self.har['log'] and self.har['log']['pages'][0]['title']:
             return self.har['log']['pages'][0]['title']
         else:
             return '!! No title found !!'
@@ -204,7 +214,7 @@ class HarFile():
         return '-'
 
     @property
-    def entries(self) -> List[Dict[str, Any]]:
+    def entries(self) -> list[dict[str, Any]]:
         """List of all the entries in the capture"""
         if not self.har['log']['entries']:
             raise Har2TreeError('Empty HAR file (no entries), invalid capture.')
@@ -215,7 +225,7 @@ class HarFile():
         """First URL of the capture"""
         return self.entries[0]['request']['url']
 
-    def __find_header_value(self, har_entry: Dict[str, Any], header_name: str) -> Optional[str]:
+    def __find_header_value(self, har_entry: dict[str, Any], header_name: str) -> str | None:
         """Get the value of a specific header"""
         for header_entry in har_entry['request']['headers']:
             if header_entry['name'].lower() == header_name.lower():
@@ -229,7 +239,7 @@ class HarFile():
         return self.entries[0]['request']['url'] != self.final_redirect
 
     @property
-    def initial_redirects(self) -> List[str]:
+    def initial_redirects(self) -> list[str]:
         '''All the initial redirects from the URL given by the user (if they can be found without building the tree)'''
         to_return = []
         if self.has_initial_redirects:
@@ -265,12 +275,12 @@ class HarFile():
         return to_return
 
     @property
-    def root_referrer(self) -> Optional[str]:
+    def root_referrer(self) -> str | None:
         '''Get the referer if the first entry. Only relevant when there are multiple tree to attach together'''
         return self.__find_header_value(self.entries[0], 'Referer')
 
     @property
-    def root_user_agent(self) -> Optional[str]:
+    def root_user_agent(self) -> str | None:
         '''Get the User Agent of the first entry'''
         return self.__find_header_value(self.entries[0], 'User-Agent')
 
@@ -289,39 +299,39 @@ class Har2Tree:
         self.har = HarFile(har_path, capture_uuid)
         self.hostname_tree = HostNode(capture_uuid=self.har.capture_uuid)
 
-        self._nodes_list: List[URLNode] = []
-        self.all_url_requests: Dict[str, List[URLNode]] = {unquote_plus(url_entry['request']['url']): [] for url_entry in self.har.entries}
+        self._nodes_list: list[URLNode] = []
+        self.all_url_requests: dict[str, list[URLNode]] = {unquote_plus(url_entry['request']['url']): [] for url_entry in self.har.entries}
 
         # Format: pageref: node UUID
-        self.pages_root: Dict[str, str] = {}
+        self.pages_root: dict[str, str] = {}
 
-        self.all_redirects: List[str] = []
-        self.all_referer: Dict[str, List[str]] = defaultdict(list)
-        self.all_initiator_url: Dict[str, List[str]] = defaultdict(list)
+        self.all_redirects: list[str] = []
+        self.all_referer: dict[str, list[str]] = defaultdict(list)
+        self.all_initiator_url: dict[str, list[str]] = defaultdict(list)
         self._load_url_entries()
 
         # Generate cookies lookup tables
         # All the initial cookies sent with the initial request given to splash
-        self.initial_cookies: Dict[str, Dict[str, Any]] = {}
+        self.initial_cookies: dict[str, dict[str, Any]] = {}
         if hasattr(self._nodes_list[0], 'cookies_sent'):
             self.initial_cookies = {key: cookie for key, cookie in self._nodes_list[0].cookies_sent.items()}
 
         # Dictionary of all cookies received during the capture
-        self.cookies_received: Dict[str, List[Tuple[str, URLNode, bool]]] = defaultdict(list)
+        self.cookies_received: dict[str, list[tuple[str, URLNode, bool]]] = defaultdict(list)
         for n in self._nodes_list:
             if hasattr(n, 'cookies_received'):
                 for domain, c_received, is_3rd_party in n.cookies_received:
                     self.cookies_received[c_received].append((domain, n, is_3rd_party))
 
         # Dictionary of all cookies sent during the capture
-        self.cookies_sent: Dict[str, List[URLNode]] = defaultdict(list)
+        self.cookies_sent: dict[str, list[URLNode]] = defaultdict(list)
         for n in self._nodes_list:
             if hasattr(n, 'cookies_sent'):
                 for c_sent in n.cookies_sent.keys():
                     self.cookies_sent[c_sent].append(n)
 
         # NOTE: locally_created contains all cookies not present in a response, and not passed at the begining of the capture to splash
-        self.locally_created: Dict[str, Dict[str, Any]] = {}
+        self.locally_created: dict[str, dict[str, Any]] = {}
         for c in self.har.cookies:
             c_identifier = f'{c["name"]}={c["value"]}'
             if (c_identifier not in self.cookies_received
@@ -329,7 +339,7 @@ class Har2Tree:
                 self.locally_created[f'{c["name"]}={c["value"]}'] = c
 
         # NOTE: locally_created_not_sent only contains cookies that are created locally, and never sent during the capture
-        self.locally_created_not_sent: Dict[str, Dict[str, Any]] = self.locally_created.copy()
+        self.locally_created_not_sent: dict[str, dict[str, Any]] = self.locally_created.copy()
         # Cross reference the source of the cookie
         for n in self._nodes_list:
             if hasattr(n, 'cookies_sent'):
@@ -389,7 +399,7 @@ class Har2Tree:
         self.url_tree = self._nodes_list.pop(0)
 
     @property
-    def initial_referer(self) -> Optional[str]:
+    def initial_referer(self) -> str | None:
         '''The referer passed to the first URL in the tree'''
         if hasattr(self.url_tree, 'referer'):
             return self.url_tree.referer
@@ -404,15 +414,15 @@ class Har2Tree:
         return sum(urlnode.body.getbuffer().nbytes for urlnode in self.url_tree.traverse() if not urlnode.empty_response)
 
     @property
-    def stats(self) -> Dict[str, Any]:
+    def stats(self) -> dict[str, Any]:
         """Statistics about the capture"""
-        to_return: Dict[str, Any] = {'total_hostnames': 0}
+        to_return: dict[str, Any] = {'total_hostnames': 0}
         to_return['total_urls'] = sum(1 for _ in self.url_tree.traverse())
         to_return['total_unique_urls'] = len({node.name for node in self.url_tree.traverse()})
 
-        all_cookies_sent: Set[str] = set()
-        all_cookies_received: Set[Tuple[str, str, bool]] = set()
-        all_hostnames: Set[str] = set()
+        all_cookies_sent: set[str] = set()
+        all_cookies_received: set[tuple[str, str, bool]] = set()
+        all_hostnames: set[str] = set()
         for host_node in self.hostname_tree.traverse():
             to_return['total_hostnames'] += 1
             all_hostnames.add(host_node.name)
@@ -430,12 +440,12 @@ class Har2Tree:
         return to_return
 
     @property
-    def redirects(self) -> List[str]:
+    def redirects(self) -> list[str]:
         """List of redirects for this tree"""
         return [a.name for a in reversed(self.rendered_node.get_ancestors())] + [self.rendered_node.name]
 
     @property
-    def root_referer(self) -> Optional[str]:
+    def root_referer(self) -> str | None:
         '''Referer if the first entry. Only relevant when there are multiple tree to attach together'''
         return self.har.root_referrer
 
@@ -450,12 +460,12 @@ class Har2Tree:
         return self.url_tree.start_time
 
     @lru_cache
-    def build_all_hashes(self, algorithm: str='sha1') -> Dict[str, List[URLNode]]:
+    def build_all_hashes(self, algorithm: str='sha1') -> dict[str, list[URLNode]]:
         '''Build on demand hashes for all the ressources of the tree, in the alorighm provided by the user'''
         if algorithm not in hashlib.algorithms_available:
             raise Har2TreeError(f'Invalid algorithm ({algorithm}), only the following are supported: {hashlib.algorithms_available}')
 
-        to_return: Dict[str, List[URLNode]] = defaultdict(list)
+        to_return: dict[str, list[URLNode]] = defaultdict(list)
         for urlnode in self.url_tree.traverse():
             if urlnode.empty_response:
                 continue
@@ -480,9 +490,9 @@ class Har2Tree:
         # loading a specific URL, and we often have an other entry requesting the same URL (but not always)
         # * If we have an other query to the same URL in the list, discard the one with a status code to 0
         # * If we don't, use the entry in the tree.
-        entries_with_0_status: Dict[str, List[int]] = defaultdict(list)
-        entries_with_negative_status: Dict[str, List[int]] = defaultdict(list)
-        ignore: List[int] = []
+        entries_with_0_status: dict[str, list[int]] = defaultdict(list)
+        entries_with_negative_status: dict[str, list[int]] = defaultdict(list)
+        ignore: list[int] = []
         for i, url_entry in enumerate(self.har.entries):
             url = unquote_plus(url_entry["request"]["url"])
             if url_entry['response']['status'] == 0:
@@ -556,7 +566,8 @@ class Har2Tree:
         if node:
             return node[0]
 
-        if self.har.final_redirect:
+        browser_errors = ['chrome-error', 'about:blank']
+        if self.har.final_redirect and not any(self.har.final_redirect.startswith(r) for r in browser_errors):
             self.logger.warning(f'Final redirect URL from adress bar not in tree: {self.har.final_redirect}')
         else:
             # No final redirect, already logged earlier.
@@ -579,28 +590,33 @@ class Har2Tree:
         """Dump the whole HostNode tree to json (for d3js)"""
         return self.hostname_tree.to_json()
 
-    def make_hostname_tree(self, root_nodes_url: Union[URLNode, List[URLNode]], root_node_hostname: HostNode) -> None:
+    def make_hostname_tree(self, root_nodes_url: URLNode | list[URLNode], root_node_hostname: HostNode) -> None:
         """ Groups all the URLs by domain in the HostNode tree.
         `root_node_url` can be a list of nodes called by the same `root_node_hostname`
         """
         if not isinstance(root_nodes_url, list):
             root_nodes_url = [root_nodes_url]
         for root_node_url in root_nodes_url:
-            children_hostnames: Dict[str, HostNode] = {}
-            sub_roots: Dict[HostNode, List[URLNode]] = defaultdict(list)
+            children_hostnames: dict[str, HostNode] = {}
+            sub_roots: dict[HostNode, list[URLNode]] = defaultdict(list)
             for child_node_url in root_node_url.get_children():
                 if child_node_url.hostname is None:
                     self.logger.warning(f'Fucked up URL: {child_node_url}')
                     continue
-                if child_node_url.hostname in children_hostnames:
-                    child_node_hostname = children_hostnames[child_node_url.hostname]
-                else:
-                    child_node_hostname = root_node_hostname.add_child(HostNode(capture_uuid=self.har.capture_uuid, name=child_node_url.hostname))
+                # Attempt to find an existing HostNode with the same hostname
+                if child_node_url.hostname not in children_hostnames:
+                    # create a new hostnode, add it in the tree
+                    child_node_hostname = HostNode(capture_uuid=self.har.capture_uuid)
+                    root_node_hostname.add_child(child_node_hostname)
                     children_hostnames[child_node_url.hostname] = child_node_hostname
+                else:
+                    child_node_hostname = children_hostnames[child_node_url.hostname]
+
                 child_node_hostname.add_url(child_node_url)
 
                 if not child_node_url.is_leaf():
                     sub_roots[child_node_hostname].append(child_node_url)
+
             for child_node_hostname, child_nodes_url in sub_roots.items():
                 self.make_hostname_tree(child_nodes_url, child_node_hostname)
 
@@ -621,10 +637,15 @@ class Har2Tree:
                 self._nodes_list = []
 
         # 2022-08-25: We now have a tree, we have a self.rendered_node, attach the features.
-        if self.har.html_content:
-            self.rendered_node.add_rendered_features(list(self.all_url_requests.keys()), rendered_html=self.har.html_content)
-        elif self.har.downloaded_file and self.har.downloaded_filename:
-            self.rendered_node.add_rendered_features(list(self.all_url_requests.keys()), downloaded_file=(self.har.downloaded_filename, self.har.downloaded_file))
+        # 2023-09-26: Turns out we can have a download *and* a rendered HTML if the JS on the page triggers a download
+        downloaded_file = None
+        if self.har.downloaded_filename and self.har.downloaded_file:
+            downloaded_file = (self.har.downloaded_filename if self.har.downloaded_filename else '',
+                               self.har.downloaded_file if self.har.downloaded_file else None)
+        if not self.rendered_node.empty_response:
+            self.rendered_node.add_rendered_features(list(self.all_url_requests.keys()),
+                                                     rendered_html=self.har.html_content if self.har.html_content else None,
+                                                     downloaded_file=downloaded_file)
 
         # Initialize the hostname tree root
         self.hostname_tree.add_url(self.url_tree)
@@ -651,10 +672,12 @@ class Har2Tree:
 
         # Sometimes, the har has a list of pages, generally when we have HTTP redirects.
         # IF we have more than one page in the list
-        # AND the orphan node's pageref points to an other page than the first one <= FIXME not enabled yet
+        # AND the orphan node's pageref points to an other page than the first one
         # AND we already have a node in the tree with this pageref
         # => attach to that node.
-        if len(self.har.har['log']['pages']) > 1 and node.pageref != self.har.har['log']['pages'][0] and self.pages_root[node.pageref] != node.uuid:
+        if ('pages' in self.har.har['log'] and len(self.har.har['log']['pages']) > 1
+                and node.pageref != self.har.har['log']['pages'][0]
+                and self.pages_root[node.pageref] != node.uuid):
             # In that case, we check if there is already a page with the pageref of the orphan node,
             # and attach the node to that. NOTE: we can only do that if there is already a node with this pageref in the tree.
             # This node is not a page root, we can attach it \o/
@@ -670,7 +693,7 @@ class Har2Tree:
             if dev_debug:
                 self.logger.warning(f'Failed to attach URLNode in the normal process, attaching node to final redirect: {self.har.final_redirect}.')
             self._make_subtree(self.url_tree.search_nodes(name=self.har.final_redirect)[0], [node])
-        else:
+        elif 'pages' in self.har.har['log']:
             # No luck, the node is root for this pageref, let's attach it to the prior page in the list, or the very first node (tree root)
             page_before = self.har.har['log']['pages'][0]
             for page in self.har.har['log']['pages'][1:]:
@@ -692,11 +715,14 @@ class Har2Tree:
                 page_root_node = self.url_tree
                 self.logger.warning('The pages in the HAR are in in the wrong order, this should not happen but here we are')
             self._make_subtree(page_root_node, [node])
+        else:
+            # no way to attach it to anything else, attach to the root node
+            self._make_subtree(self.url_tree, [node])
 
     @trace_make_subtree
-    def _make_subtree(self, root: URLNode, nodes_to_attach: Optional[List[URLNode]]=None, dev_debug: bool=False) -> None:
+    def _make_subtree(self, root: URLNode, nodes_to_attach: list[URLNode] | None=None, dev_debug: bool=False) -> None:
         """Recursive method building each level of the tree"""
-        matching_urls: List[URLNode]
+        matching_urls: list[URLNode]
         if nodes_to_attach is None:
             # We're in the actual root node
             unodes = [self.url_tree]
@@ -705,7 +731,6 @@ class Har2Tree:
             for unode in nodes_to_attach:
                 if dev_debug:
                     self.logger.warning(f'Attaching URLNode {unode.name} to {root.name}.')
-
                 unodes.append(root.add_child(unode))
 
         if dev_debug:
@@ -751,8 +776,13 @@ class Har2Tree:
                 else:
                     self.logger.warning(f'The URLNode has a redirect to something we already processed ({unode.redirect_url}), this should not happen.')
 
-            # The node can have a redirect, but also trigger ressources refering to themselves, we need to trigger this code on each node.
+            # 2025-02-06: If a node has no redirect **and** no content (empty response), we don't want to attach anything to it (it is a leaf)
+            # Example: A POST to self that triggers the **parent** to load an other URL. In this case,
+            # the proper attachment point is the parent, not this node, even if we have other nodes with this node URL as a referer.
+            if unode.empty_response:
+                continue
 
+            # The node can have a redirect, but also trigger ressources refering to themselves, we need to trigger this code on each node.
             if self.all_initiator_url.get(unode.name):
                 # The URL (unode.name) is in the list of known urls initiating calls
                 for u in self.all_initiator_url[unode.name]:
@@ -765,21 +795,20 @@ class Har2Tree:
 
             # 2022-04-27: Referers are supposed to be the complete URL, but they can also be only part of it.
             #             what we do here is going from best to worse in the hope one of them matches
-            _referer_strings: List[str] = [
+            _referer_strings: list[str] = [
                 unode.name,  # full URL
                 unode.alternative_url_for_referer,  # URL up to the first `#`
             ]
             for ref in _referer_strings:
                 if self.all_referer.get(ref):
                     matching_urls = []
+                    # 2024-11-20: Referers are kinda weak, we can have multiple URLs with the same referer even one of the nodes should be attached somewhere else.
+                    #             Let's attach the nodes one by one there (if they've not been attached recursively)
                     for u in self.all_referer[ref]:
-                        matching_urls += [url_node for url_node in self.all_url_requests[u]
-                                          if url_node in self._nodes_list and hasattr(url_node, 'referer') and url_node.referer == ref]
-                        self._nodes_list = [node for node in self._nodes_list if node not in matching_urls]
-                        if dev_debug:
-                            self.logger.warning(f'Found via referer from {unode.name} to {matching_urls}.')
-                    # 2022-04-27: build subtrees recursively *after* we find all the best referer matches
-                    self._make_subtree(unode, matching_urls)
+                        for url_node in self.all_url_requests[u]:
+                            if url_node in self._nodes_list and hasattr(url_node, 'referer') and url_node.referer == ref:
+                                self._nodes_list = [node for node in self._nodes_list if node != url_node]
+                                self._make_subtree(unode, [url_node])
 
             if hasattr(unode, 'external_ressources'):
                 # the url loads external things, and some of them have no referer....
@@ -789,7 +818,13 @@ class Har2Tree:
                             # We have a lot of false positives
                             # 2021-06-19: or the URL of the final redirect is somewhere in an embeded content. In that case, we don't want to attach to the sub-node.
                             continue
-                        matching_urls = [url_node for url_node in self.all_url_requests[link] if url_node in self._nodes_list]
+                        # 2024-11-21: We only want to attach via external ressources *if*
+                        # 1. the node has no referer
+                        # 2. the node has a referer, but it is not in the list of potental nodes we can attach them to so they would be dangling.
+
+                        matching_urls = [url_node for url_node in self.all_url_requests[link]
+                                         if url_node in self._nodes_list
+                                         and (not hasattr(url_node, 'referer') or url_node.referer not in self.all_url_requests)]
                         self._nodes_list = [node for node in self._nodes_list if node not in matching_urls]
                         if dev_debug:
                             self.logger.warning(f'Found from {unode.name} via external ressources ({external_tag}): {matching_urls}.')
