@@ -1,21 +1,25 @@
 #!/usr/bin/env python
 
-from pathlib import Path
-from typing import List, Optional, Union, Tuple, Set, MutableMapping, Any, Dict
+from __future__ import annotations
+
+import binascii
+import copy
+import hashlib
+import ipaddress
+import json
 import logging
 import uuid
-import json
-from .helper import find_external_ressources, rebuild_url
-from io import BytesIO
-from urllib.parse import unquote_plus, urlparse, urljoin
-from datetime import datetime, timedelta
-import ipaddress
-from base64 import b64decode
-import hashlib
 import re
-from functools import lru_cache
 
-from .helper import Har2TreeError, Har2TreeLogAdapter, make_hhhash, HHHashError
+from base64 import b64decode
+from datetime import datetime, timedelta
+from functools import lru_cache, cached_property
+from hashlib import sha256
+from io import BytesIO
+from pathlib import Path
+from typing import Any
+from collections.abc import MutableMapping
+from urllib.parse import unquote_plus, urlparse, urljoin
 
 import filetype  # type: ignore
 from bs4 import BeautifulSoup
@@ -24,6 +28,9 @@ from publicsuffixlist import PublicSuffixList  # type: ignore
 from w3lib.html import strip_html5_whitespace
 from w3lib.url import canonicalize_url, safe_url_string
 
+from .helper import find_external_ressources, rebuild_url, find_identifiers, make_soup
+from .helper import Har2TreeError, Har2TreeLogAdapter, make_hhhash, HHHashError, HHHashNote
+
 
 @lru_cache(64)
 def get_public_suffix_list() -> PublicSuffixList:
@@ -32,7 +39,7 @@ def get_public_suffix_list() -> PublicSuffixList:
     return PublicSuffixList()
 
 
-class HarTreeNode(TreeNode):
+class HarTreeNode(TreeNode):  # type: ignore[misc]
 
     def __init__(self, capture_uuid: str, **kwargs: Any):
         """Node dumpable in json to display with d3js"""
@@ -63,6 +70,8 @@ class HarTreeNode(TreeNode):
 
 class URLNode(HarTreeNode):
 
+    start_time: datetime
+
     def __init__(self, capture_uuid: str, **kwargs: Any):
         """Node of the URL Tree"""
         super().__init__(capture_uuid=capture_uuid, **kwargs)
@@ -74,28 +83,55 @@ class URLNode(HarTreeNode):
         self.features_to_skip.add('time_content_received')
         self.features_to_skip.add('ip_address')
 
-    def add_rendered_features(self, all_requests: List[str], rendered_html: Optional[BytesIO]=None, downloaded_file: Tuple[str, Optional[BytesIO]]=('', None)) -> None:
+    def _compute_domhash(self) -> str:
+        to_hash = "|".join(t.name for t in self.rendered_soup.find_all()).encode()
+        return sha256(to_hash).hexdigest()[:32]
+
+    def add_rendered_features(self, all_requests: list[str], rendered_html: BytesIO | None=None, downloaded_file: tuple[str, BytesIO | None] | None=None) -> None:
         if rendered_html:
             self.add_feature('rendered_html', rendered_html)
-            rendered_external, rendered_embedded = find_external_ressources(rendered_html.getvalue(), self.name, all_requests)
+            rendered_external, rendered_embedded = find_external_ressources(self.mimetype, self.rendered_html.getvalue(), self.name, all_requests)
             if hasattr(self, 'external_ressources'):
                 # for the external ressources, the keys are always the same
-                self.external_ressources: Dict[str, List[str]] = {initiator_type: urls + rendered_external[initiator_type] for initiator_type, urls in self.external_ressources.items()}
+                self.external_ressources: dict[str, list[str]] = {initiator_type: urls + rendered_external[initiator_type] for initiator_type, urls in self.external_ressources.items()}
             else:
                 self.add_feature('external_ressources', rendered_external)
 
             if hasattr(self, 'embedded_ressources'):
                 # for the embedded ressources, the keys are the mimetypes, they may not overlap
                 mimetypes = list(self.embedded_ressources.keys()) + list(rendered_embedded.keys())
-                self.embedded_ressources: Dict[str, List[Tuple[str, BytesIO]]] = {mimetype: self.embedded_ressources.get(mimetype, []) + rendered_embedded.get(mimetype, []) for mimetype in mimetypes}
+                self.embedded_ressources: dict[str, list[tuple[str, BytesIO]]] = {mimetype: self.embedded_ressources.get(mimetype, []) + rendered_embedded.get(mimetype, []) for mimetype in mimetypes}
             else:
                 self.add_feature('embedded_ressources', rendered_embedded)
-        elif downloaded_file:
+
+            if identifiers := find_identifiers(self.rendered_soup):
+                self.add_feature('identifiers', identifiers)
+
+            if domhash := self._compute_domhash():
+                self.add_feature('domhash', domhash)
+
+        if downloaded_file:
             downloaded_filename, downloaded_file_data = downloaded_file
             self.add_feature('downloaded_file', downloaded_file_data)
             self.add_feature('downloaded_filename', downloaded_filename)
 
-    def load_har_entry(self, har_entry: MutableMapping[str, Any], all_requests: List[str]) -> None:
+    def _dirty_safe_b64decode(self, to_decode: str | bytes) -> bytes:
+        if isinstance(to_decode, str):
+            # make it bytes
+            _to_decode = to_decode.encode()
+        else:
+            # Make sure we're not modifying it
+            _to_decode = copy.copy(to_decode)
+        # Make sure the padding is right
+        if len(_to_decode) % 4:
+            _to_decode += b'==='
+
+        if any(altchar in _to_decode for altchar in [b'-', b'_']):
+            # Emulate urlsafe b64decode
+            return b64decode(_to_decode, altchars=b'-_', validate=True)
+        return b64decode(_to_decode, validate=True)
+
+    def load_har_entry(self, har_entry: MutableMapping[str, Any], all_requests: list[str]) -> None:
         """Load one entry of the HAR file, initialize most of the features of the node"""
         if not self.name:
             # We're in the actual root node
@@ -148,6 +184,15 @@ class URLNode(HarTreeNode):
             # Not an IP
             pass
 
+        if not hasattr(self, 'hostname_is_ip'):
+            try:
+                # attempt to decode if the hostname is idna encoded
+                idna_decoded = self.hostname.encode().decode('idna')
+                if idna_decoded != self.hostname:
+                    self.add_feature('idna', idna_decoded)
+            except UnicodeError:
+                pass
+
         if not hasattr(self, 'hostname_is_ip') and not hasattr(self, 'file_on_disk'):
             tld = get_public_suffix_list().publicsuffix(self.hostname)
             if tld:
@@ -166,85 +211,123 @@ class URLNode(HarTreeNode):
         if 'user_agent' not in self.features:
             self.add_feature('user_agent', '')
 
-        if 'method' in self.request and self.request['method'] == 'POST' and 'postData' in self.request:
-            # If the content is empty, we don't care
-            if self.request['postData']['text']:
-                # We have a POST request, the data can be base64 encoded or urlencoded
-                posted_data: Union[str, bytes] = self.request['postData']['text']
-                if 'encoding' in self.request['postData']:
-                    if self.request['postData']['encoding'] == 'base64':
-                        if len(posted_data) % 4:
-                            # a this point, we have a string for sure
-                            posted_data += '==='  # type: ignore
-                        posted_data = b64decode(posted_data)
-                    else:
-                        self.logger.warning(f'Unexpected encoding: {self.request["postData"]["encoding"]}')
-
-                if 'mimeType' in self.request['postData']:
-                    if self.request['postData']['mimeType'].startswith('application/x-www-form-urlencoded'):
-                        # 100% sure there will be websites where decode will fail
-                        if isinstance(posted_data, bytes):
-                            try:
-                                posted_data = posted_data.decode()
-                            except Exception:
-                                self.logger.debug(f'Expected urlencoded, got garbage: {posted_data!r}')
-                        if isinstance(posted_data, str):
-                            posted_data = unquote_plus(posted_data)
-                    elif (self.request['postData']['mimeType'].startswith('application/json')
-                          or self.request['postData']['mimeType'].startswith('application/csp-report')
-                          or self.request['postData']['mimeType'].startswith('application/x-amz-json-1.1')
-                          or self.request['postData']['mimeType'].startswith('application/x-json-stream')
-                          or self.request['postData']['mimeType'].startswith('application/reports+json')
-                          ):
-                        try:
-                            # it is relatively common that the supposedly json blob is base64 encoded but the encoding wasn't set properly, let's try to decode anyway.
-                            posted_data = b64decode(posted_data)
-                            self.logger.info("Contains a sneakily base64 encoded json blob.")
-                        except Exception:
-                            pass
-
-                        if posted_data:
-                            try:
-                                posted_data = json.loads(posted_data)
-                            except Exception:
-                                self.logger.debug(f"Expected json, got garbage: {self.request['postData']['mimeType']} - {posted_data[:20]!r}[...]")
-
-                    elif self.request['postData']['mimeType'].startswith('multipart/form-data'):
-                        # FIXME multipart content (similar to email). Not totally sure what do do with it tight now.
-                        pass
-                    elif self.request['postData']['mimeType'].startswith('application/x-protobuffer'):
-                        # FIXME If possible, decode?
-                        pass
-                    elif self.request['postData']['mimeType'].startswith('text'):
-                        # We got text, keep what we already have
-                        pass
-                    elif self.request['postData']['mimeType'] == '?':
-                        # Just skip it, no need to go in the warnings
-                        pass
-                    elif self.request['postData']['mimeType'] in ['application/octet-stream']:
-                        # Should flag it.
-                        pass
-                    else:
-                        # Weird stuff: Image/GIF application/unknown application/grpc-web+proto
-                        self.logger.warning(f'Unexpected mime type: {self.request["postData"]["mimeType"]}')
-
-                # The data may be json, try to load it
+        if 'method' in self.request and self.request['method'] == 'POST':
+            decoded_posted_data: str | bytes | int | float | bool | dict[str, str] | None = None
+            if 'postData' not in self.request or 'text' not in self.request['postData']:
+                self.logger.debug('POST request with no content.')
+            elif not self.request['postData']['text']:
+                # If the POST content is empty
+                self.logger.debug('Empty POST request.')
+                decoded_posted_data = ''
+            elif self.request['postData'].get('params'):
+                # NOTE 2025-08-08
+                # if the posted data mimetype is "application/x-www-form-urlencoded"
+                # the HAR contains the decoded entry in the params key
+                # The params key is a list of dicts with a key and a value
+                #   {"name": <key>, "value": <data>}
+                # I'd rather have it as {<key>: <data>}
+                # TODO: some processing on the data part (it's often a json blob)
+                self.logger.debug('Got a params POST.')
+                decoded_posted_data = {entry['name']: entry['value'] for entry in self.request['postData']['params']}
+            else:
+                # NOTE 2023-08-22: Blind attempt to base64 decode the data
+                self.logger.debug('Got a normal POST')
                 try:
-                    posted_data = json.loads(posted_data)
+                    decoded_posted_data = self._dirty_safe_b64decode(self.request['postData']['text'])
+                except binascii.Error:
+                    decoded_posted_data = self.request['postData']['text']
+                if 'mimeType' in self.request['postData']:
+                    # make it easier to compare.
+                    mimetype_lower = self.request['postData']['mimeType'].lower()
+                    if mimetype_lower.startswith('application/x-www-form-urlencoded'):
+                        # NOTE: this should never happen as there should
+                        # be something in self.request['postData']['params']
+                        # and we already processed it before but just in case...
+                        self.logger.warning(f'Got a application/x-www-form-urlencoded without params key: {self.request}')
+                        # 100% sure there will be websites where decode will fail
+                        try:
+                            if isinstance(decoded_posted_data, bytes):
+                                decoded_posted_data = decoded_posted_data.decode()
+                            if isinstance(decoded_posted_data, str):
+                                decoded_posted_data = unquote_plus(decoded_posted_data)
+                        except Exception as e:
+                            self.logger.warning(f'Unable to unquote form data "{decoded_posted_data!r}": {e}')
+                    elif (mimetype_lower.startswith('application/json')
+                          or mimetype_lower.startswith('application/csp-report')
+                          or mimetype_lower.startswith('application/x-amz-json-1.1')
+                          or mimetype_lower.startswith('application/x-json-stream')
+                          or mimetype_lower.startswith('application/reports+json')
+                          or mimetype_lower.endswith('json')
+                          ):
+                        if isinstance(decoded_posted_data, (str, bytes)):
+                            # at this stage, it will always be bytes or str
+                            try:
+                                # NOTE 2023-08-22: loads here may give us a int, float or a bool.
+                                decoded_posted_data = json.loads(decoded_posted_data)
+                            except Exception:
+                                if isinstance(decoded_posted_data, (str, bytes)):
+                                    self.logger.warning(f"Expected json, got garbage: {mimetype_lower} - {decoded_posted_data[:20]!r}[...]")
+                                else:
+                                    self.logger.warning(f"Expected json, got garbage: {mimetype_lower} - {decoded_posted_data}")
+
+                    elif mimetype_lower.startswith('multipart/form-data'):
+                        # FIXME multipart content (similar to email). Not totally sure what do do with it tight now.
+                        self.logger.debug(f'Got a POST {mimetype_lower}: {decoded_posted_data!r}')
+                        pass
+                    elif mimetype_lower.startswith('application/x-protobuf'):
+                        # FIXME If possible, decode?
+                        self.logger.debug(f'Got a POST {mimetype_lower}: {decoded_posted_data!r}')
+                        pass
+                    elif mimetype_lower.startswith('text') and isinstance(decoded_posted_data, (str, bytes)):
+                        try:
+                            # NOTE 2023-08-22: Quite a few text entries are in fact json, give it a shot.
+                            # loads here may give us a int, float or a bool.
+                            decoded_posted_data = json.loads(decoded_posted_data)
+                        except Exception:
+                            # keep it as it is otherwise.
+                            pass
+                    elif mimetype_lower.endswith('javascript'):
+                        # keep it as it is
+                        self.logger.warning(f'Got a POST {mimetype_lower}: {decoded_posted_data!r}')
+                        pass
+                    elif mimetype_lower == '?':
+                        # Just skip it, no need to go in the warnings
+                        self.logger.warning(f'Got a POST {mimetype_lower}: {decoded_posted_data!r}')
+                        pass
+                    elif mimetype_lower in ['application/octet-stream', 'application/binary']:
+                        # Should flag it, maybe?
+                        self.logger.warning(f'Got a POST {mimetype_lower}: {decoded_posted_data!r}')
+                        pass
+                    elif mimetype_lower in ['application/unknown', 'application/grpc-web+proto']:
+                        # Weird but already seen stuff
+                        self.logger.warning(f'Got a POST {mimetype_lower}: {decoded_posted_data!r}')
+                        pass
+                    else:
+                        self.logger.warning(f'Unexpected mime type: {mimetype_lower}')
+                else:
+                    self.logger.warning(f'Missing mimetype in POST: {self.request["postData"]}')
+
+            # NOTE 2023-08-22: Blind attempt to process the data as json
+            if isinstance(decoded_posted_data, (str, bytes)):
+                try:
+                    decoded_posted_data = json.loads(decoded_posted_data)
                 except Exception:
                     pass
 
-                if isinstance(posted_data, bytes):
-                    # Try to decode it as utf-8
-                    try:
-                        posted_data = posted_data.decode('utf-8')
-                    except Exception:
-                        pass
-                self.add_feature('posted_data', posted_data)
+            if isinstance(decoded_posted_data, bytes):
+                # NOTE 2023-08-22: Blind attempt to decode the bytes
+                # Try to decode it as utf-8
+                try:
+                    decoded_posted_data = decoded_posted_data.decode('utf-8')
+                except Exception:
+                    pass
+            self.add_feature('posted_data', decoded_posted_data)
 
         self.add_feature('response', har_entry['response'])
         try:
             self.add_feature('hhhash', make_hhhash(self.response))
+        except HHHashNote as e:
+            self.logger.debug(e)
         except HHHashError as e:
             self.logger.warning(e)
 
@@ -284,10 +367,14 @@ class URLNode(HarTreeNode):
         if not self.response['content'].get('text') or self.response['content']['text'] == '':
             # If the content of the response is empty, skip.
             self.add_feature('empty_response', True)
+            self.add_feature('mimetype', 'inode/x-empty')
         else:
             self.add_feature('empty_response', False)
             if self.response['content'].get('encoding') == 'base64':
-                self.add_feature('body', BytesIO(b64decode(self.response['content']['text'])))
+                try:
+                    self.add_feature('body', BytesIO(self._dirty_safe_b64decode(self.response['content']['text'])))
+                except binascii.Error:
+                    self.add_feature('body', BytesIO(self.response['content']['text'].encode()))
             else:
                 self.add_feature('body', BytesIO(self.response['content']['text'].encode()))
 
@@ -305,7 +392,7 @@ class URLNode(HarTreeNode):
             if not hasattr(self, 'mimetype'):
                 self.add_feature('mimetype', '')
 
-            external_ressources, embedded_ressources = find_external_ressources(self.body.getvalue(), self.name, all_requests)
+            external_ressources, embedded_ressources = find_external_ressources(self.mimetype, self.body.getvalue(), self.name, all_requests)
             self.add_feature('external_ressources', external_ressources)
             self.add_feature('embedded_ressources', embedded_ressources)
 
@@ -333,12 +420,6 @@ class URLNode(HarTreeNode):
                     self.add_feature('redirect', True)
                     self.add_feature('redirect_url', self.external_ressources['meta_refresh'][0])
 
-            # FIXME: Deprecated, use generic_type directly. Keep for now for backward compat
-            self.add_feature(self.generic_type, True)
-            if self.generic_type == 'unknown_mimetype':
-                if self.mimetype not in ['x-unknown']:
-                    self.logger.warning(f'Unknown mimetype: {self.mimetype}')
-
         # NOTE: Chrome/Chromium/Playwright only feature
         if har_entry.get('serverIPAddress'):
             # check ipv6 format
@@ -390,43 +471,7 @@ class URLNode(HarTreeNode):
                     original_redirect=self.response['redirectURL'],
                     modified_redirect=redirect_url))
 
-    @property
-    def generic_type(self) -> str:
-        if 'javascript' in self.mimetype or 'ecmascript' in self.mimetype:
-            return 'js'
-        elif self.mimetype.startswith('image'):
-            return 'image'
-        elif self.mimetype.startswith('text/css'):
-            return 'css'
-        elif 'json' in self.mimetype:
-            return 'json'
-        elif 'html' in self.mimetype:
-            return 'html'
-        elif 'font' in self.mimetype or 'woff' in self.mimetype:
-            return 'font'
-        elif 'octet-stream' in self.mimetype:
-            return 'octet-stream'
-        elif ('text/plain' in self.mimetype or 'xml' in self.mimetype
-                or 'application/x-www-form-urlencoded' in self.mimetype
-                or 'application/vnd.oasis.opendocument.formula-template' in self.mimetype):
-            return 'text'
-        elif 'video' in self.mimetype:
-            return 'video'
-        elif 'audio' in self.mimetype:
-            return 'audio'
-        elif 'mpegurl' in self.mimetype.lower():
-            return 'livestream'
-        elif ('application/x-shockwave-flash' in self.mimetype
-                or 'application/x-shockware-flash' in self.mimetype):  # Yes, shockwaRe
-            return 'flash'
-        elif 'application/pdf' in self.mimetype:
-            return 'pdf'
-        elif not self.mimetype:
-            return 'unset_mimetype'
-        else:
-            return 'unknown_mimetype'
-
-    def _find_initiator_in_stack(self, stack: MutableMapping[str, Any]) -> Optional[str]:
+    def _find_initiator_in_stack(self, stack: MutableMapping[str, Any]) -> str | None:
         # Because everything is terrible, and the call stack can have parents
         if stack['callFrames']:
             return unquote_plus(stack['callFrames'][0]['url'])
@@ -435,7 +480,7 @@ class URLNode(HarTreeNode):
         return None
 
     @property
-    def resources_hashes(self) -> Set[str]:
+    def resources_hashes(self) -> set[str]:
         all_ressources_hashes = set()
         if 'body_hash' in self.features:
             all_ressources_hashes.add(self.body_hash)
@@ -445,9 +490,9 @@ class URLNode(HarTreeNode):
         return all_ressources_hashes
 
     @property
-    def urls_in_rendered_page(self) -> List[str]:
+    def urls_in_rendered_page(self) -> list[str]:
 
-        def _sanitize(maybe_url: str) -> Optional[str]:
+        def _sanitize(maybe_url: str) -> str | None:
             href = strip_html5_whitespace(maybe_url)
             href = safe_url_string(href)
 
@@ -461,11 +506,10 @@ class URLNode(HarTreeNode):
 
         if not hasattr(self, 'rendered_html') or not self.rendered_html:
             raise Har2TreeError('Not the node of a page rendered, invalid request.')
-        urls: Set[str] = set()
-        soup = BeautifulSoup(self.rendered_html.getvalue(), "lxml")
+        urls: set[str] = set()
 
         # The simple ones: the links.
-        for a_tag in soup.find_all(["a", "area"]):
+        for a_tag in self.rendered_soup.find_all(["a", "area"]):
             href = a_tag.attrs.get("href")
             if not href:
                 continue
@@ -473,7 +517,7 @@ class URLNode(HarTreeNode):
                 urls.add(href)
 
         # The rest of the mess
-        for tag in soup.find_all(True):
+        for tag in self.rendered_soup.find_all(True):
             if tag.name in ["a", "area", 'img', 'script', 'video', 'audio', 'iframe', 'embed',
                             'source', 'link', 'object']:
                 # processed either above or as external resources
@@ -487,6 +531,12 @@ class URLNode(HarTreeNode):
 
         return sorted(urls)
 
+    @cached_property
+    def rendered_soup(self) -> BeautifulSoup:
+        if not hasattr(self, 'rendered_html') or not self.rendered_html:
+            raise Har2TreeError('Not the node of a page rendered, invalid request.')
+        return make_soup(self.rendered_html.getvalue())
+
 
 class HostNode(HarTreeNode):
 
@@ -497,26 +547,11 @@ class HostNode(HarTreeNode):
         self.features_to_skip.add('urls')
 
         self.add_feature('urls', [])
-        self.add_feature('js', 0)
-        self.add_feature('redirect', 0)
-        self.add_feature('redirect_to_nothing', 0)
-        self.add_feature('image', 0)
-        self.add_feature('css', 0)
-        self.add_feature('json', 0)
-        self.add_feature('html', 0)
-        self.add_feature('font', 0)
-        self.add_feature('octet_stream', 0)
-        self.add_feature('text', 0)
-        self.add_feature('pdf', 0)
-        self.add_feature('video', 0)
-        self.add_feature('unset_mimetype', 0)
-        self.add_feature('unknown_mimetype', 0)
-        self.add_feature('iframe', 0)
         self.add_feature('http_content', False)
         self.add_feature('https_content', False)
         self.add_feature('contains_rendered_urlnode', False)
-        self.cookies_sent: Set[str] = set()
-        self.cookies_received: Set[Tuple[str, str, bool]] = set()
+        self.cookies_sent: set[str] = set()
+        self.cookies_received: set[tuple[str, str, bool]] = set()
 
     def to_dict(self) -> MutableMapping[str, Any]:
         """Make a dictionary that is json dumpable for d3js"""
@@ -558,8 +593,9 @@ class HostNode(HarTreeNode):
     def add_url(self, url: URLNode) -> None:
         """Add a URL node to the Host node, initialize/update the features"""
         if not self.name:
-            # Only used when initializing the root node
             self.add_feature('name', url.hostname)
+            if hasattr(url, 'idna'):
+                self.add_feature('idna', url.idna)
 
         if hasattr(url, 'hostname_is_ip') and url.hostname_is_ip:
             self.add_feature('hostname_is_ip', True)
@@ -581,41 +617,12 @@ class HostNode(HarTreeNode):
             # Keep a set of cookies received: different URLs will receive the same cookie
             self.cookies_received.update({(domain, cookie, is_3rd_party)
                                           for domain, cookie, is_3rd_party in url.cookies_received})
-        if hasattr(url, 'js'):
-            self.js += 1
-        if hasattr(url, 'redirect'):
-            self.redirect += 1
-        if hasattr(url, 'redirect_to_nothing'):
-            self.redirect_to_nothing += 1
-        if hasattr(url, 'image'):
-            self.image += 1
-        if hasattr(url, 'css'):
-            self.css += 1
-        if hasattr(url, 'json'):
-            self.json += 1
-        if hasattr(url, 'html'):
-            self.html += 1
-        if hasattr(url, 'font'):
-            self.font += 1
-        if hasattr(url, 'octet_stream'):
-            self.octet_stream += 1
-        if hasattr(url, 'text'):
-            self.text += 1
-        if hasattr(url, 'pdf'):
-            self.pdf += 1  # FIXME: need icon
-        if hasattr(url, 'video') or hasattr(url, 'livestream') or hasattr(url, 'audio') or hasattr(url, 'flash'):
-            self.video += 1
-        if hasattr(url, 'unknown_mimetype') or hasattr(url, 'unset_mimetype'):
-            self.unknown_mimetype += 1
-        if hasattr(url, 'iframe'):
-            self.iframe += 1
-
         if url.name.startswith('http://'):
             self.http_content = True
         elif url.name.startswith('https://'):
             self.https_content = True
 
 
-def harnode_json_default(obj: 'HarTreeNode') -> MutableMapping[str, Any]:
+def harnode_json_default(obj: HarTreeNode) -> MutableMapping[str, Any]:
     if isinstance(obj, HarTreeNode):
         return obj.to_dict()
